#pragma once

#include "NWindowsVersion.h"
#include <memory.h>
#include <stdlib.h>
#include "NMathDefs.h"

const int HASHED_STRING_TABLE_SIZE = 1024;
const int MAX_HASHED_STRING_LENGTH = 256;

const int ID_HASH_BITS = 10;
const int ID_INSTANCE_BITS = 22;
const int ID_HIGH_10_MASK = 0xffc00000;
const int ID_LOW_10_MASK = 0x000003ff;
const int ID_HIGH_22_MASK = 0xfffffc00;
const int ID_LOW_22_MASK = 0x003fffff;

#define MAKESTRING(x) static const NHashedString x(#x)
#define MAKESTRINGWITH(x,y) static const NHashedString x(y)
#define MAKESTRINGSAFE(pre,x) static const NHashedString pre##x(#x)

class NHashedStringTable;

class NHashedString
{
public:
	NHashedString();

	// Made explicit so I can catch cases where char*s are being passed to functions that take NHashedStrings.
	// Not sure it's going to stay like this forever.  It's kind of nice to have that implicit conversion.
	explicit NHashedString(const char* string, bool bCaseSensitive=false);

	inline bool operator ==(const NHashedString& other) const {return (id == other.getID());}
	inline bool operator !=(const NHashedString& other) const {return (id != other.getID());}
	inline bool operator ^(const NHashedString& other) const {return (NHashedString(getString()) == NHashedString(other.getString()));}		// Force case-insensitive test. Slow!
	inline bool operator <(const NHashedString& other) const {return (id < other.getID());}
	inline bool operator <=(const NHashedString& other) const {return (id <= other.getID());}
	inline bool operator >(const NHashedString& other) const {return (id > other.getID());}
	inline bool operator >=(const NHashedString& other) const {return (id >= other.getID());}

	inline int getID() const {return id;}
	const char* getString() const;

private:

	int id;

#ifdef _DEBUG
	// For debugging ONLY, include a link to the string. And I guess the case-sensitivity, too.
	const char* DEBUG_string;
	bool DEBUG_bCaseSensitive;
#endif
};

extern NHashedString gNullStr;

class NHashedStringEntry
{
public:
	NHashedStringEntry(const char* _string);
	NHashedStringEntry() : next(NULL), string(NULL) {}
	~NHashedStringEntry();

	NHashedStringEntry* next;

	const char* getString() {return string;}

private:

	// 10/05/2009: This used to be statically allocated.  Now it's dynamic up to MAX_HASHED_STRING_LENGTH.
	char* string;
};

class NHashedStringTable
{
public:
	~NHashedStringTable();

	static NHashedStringTable* GetInstance();
	static void DestroyInstance();

	int add(int index, const char* string, bool bCaseSensitive=false);		// returns the depth within this list
	const char* getString(int index, int instance);

	void ReportStats(int& ArraySize, int& StringCount, int& RowMin, int& RowMax, float& Mean, float& Variance, float& StdDev);

private:
	NHashedStringTable();
	static NHashedStringTable* Instance;

	HANDLE hHashTableMutex;

	NHashedStringEntry* entryArray[HASHED_STRING_TABLE_SIZE];

	void recCleanup(NHashedStringEntry* entry);
};
