#include "NImage.h"
#include "NMathDefs.h"
#include "NText.h"
#include "NUtilDefs.h"

NImage::NImage(const NImage& Other)
{
	init(Other.width, Other.height, Other.channels);
	NBlitter::Blit(*this, Other, 0, 0, 0, 0, width, height, false);
}

void NImage::operator =(const NImage& Other)
{
	init(Other.width, Other.height, Other.channels);
	NBlitter::Blit(*this, Other, 0, 0, 0, 0, width, height, false);
}

bool NImage::loadBMP(const char* filename)
{
	FILE* file;
	errno_t err = fopen_s(&file, filename, "rb");

	if (err != 0)
	{
		return false;
	}

	fread(&fileheader, sizeof(BITMAPFILEHEADER), 1, file);

	memcmp(&fileheader.bfType, "BM", 2);

	fread(&infoheader, sizeof(BITMAPINFOHEADER), 1, file);

	if (infoheader.biSize != sizeof(infoheader))
	{
		return false;
	}

	if (infoheader.biPlanes != 1)
	{
		return false;
	}

	if (infoheader.biCompression != BI_RGB)
	{
		return false;
	}

	// Paint Shop Pro apparently outputs BMP files with biBitCount == 0.  Heh.
	// I should probably fix this.
	if (!init(infoheader.biWidth, infoheader.biHeight, infoheader.biBitCount / 8, infoheader.biBitCount))
	{
		return false;
	}

	// Not sure this condition is correct anymore in general, but it seems to work even for 16-color bitmaps...
	if (infoheader.biBitCount == 8)
	{
		fread(palette, sizeof(RGBQUAD) * NumColorsInPalette, 1, file);
	}

	if (NumColorsInPalette == 256)
	{
		fread(pixels, sizeof(unsigned char) * stride * height, 1, file);
	}
	else
	{
		// Assume it's 4bpp for now, durr.
		// Also probably gonna make some dangerous assumptions about stride here...
		for (int i = 0; i < height; ++i)
		{
			for (int j = 0; j < stride; j += 2)
			{
				uchar InByte = 0;
				fread(&InByte, sizeof(uchar), 1, file);

				uchar HiFour = (InByte & 0xf0) >> 4;
				uchar LoFour = (InByte & 0x0f);

				pixels[(i * stride) + j] = HiFour;
				pixels[(i * stride) + j + 1] = LoFour;
			}
		}
	}

	fclose(file);

	return true;
}

bool NImage::saveBMP(const char* filename)
{
	FILE* file;

	fopen_s(&file, filename, "wb");

	fwrite(&fileheader, sizeof(BITMAPFILEHEADER), 1, file);

	memcmp(&fileheader.bfType, "BM", 2);

	fwrite(&infoheader, sizeof(BITMAPINFOHEADER), 1, file);

	if (infoheader.biBitCount == 8)
	{
		fwrite(palette, sizeof(RGBQUAD) * NumColorsInPalette, 1, file);
	}

	fwrite(pixels, sizeof(unsigned char) * stride * height, 1, file);

	fclose(file);
	return true;
}

void NImage::loadPAL(const char* filename)
{
	FILE* file;
	errno_t err = fopen_s(&file, filename, "rb");

	if (err != 0)
	{
		return;
	}

	int Riff = 0;
	int RiffLen = 0;
	int Pal = 0;
	int Data = 0;
	int DataSize = 0;
	short Version = 0;

	fread_s(&Riff, sizeof(Riff), sizeof(Riff), 1, file);
	fread_s(&RiffLen, sizeof(RiffLen), sizeof(RiffLen), 1, file);
	fread_s(&Pal, sizeof(Pal), sizeof(Pal), 1, file);
	fread_s(&Data, sizeof(Data), sizeof(Data), 1, file);
	fread_s(&DataSize, sizeof(DataSize), sizeof(DataSize), 1, file);
	fread_s(&Version, sizeof(Version), sizeof(Version), 1, file);

	assert(Riff == 'FFIR');
	assert(Pal == ' LAP');
	assert(Data == 'atad');

	// I'm guessing at the rest of this.
	short NumEntries = 0;
	fread_s(&NumEntries, sizeof(NumEntries), sizeof(NumEntries), 1, file);

	if (palette && NumColorsInPalette == NumEntries)
	{
		fread_s(palette, sizeof(RGBQUAD) * NumEntries, sizeof(RGBQUAD), NumEntries, file);

		// These are in a different order. RGBQUAD is BGRr, palette is RGBA.
		for (int i = 0; i < NumEntries; ++i)
		{
			palette[i].rgbReserved = palette[i].rgbRed;
			palette[i].rgbRed = palette[i].rgbBlue;
			palette[i].rgbBlue = palette[i].rgbReserved;
			palette[i].rgbReserved = 0;
		}
	}
	else
	{
		UNIMPLEMENTED;
	}

	fclose(file);
}

NColor8 NImage::getPixel(int x, int y) const
{
	// This may slow stuff down, but I'm sick of getting crashes here...
	if (x < 0 || x >= width || y < 0 || y >= height)
	{
		return NColor8(0,0,0,0);
	}

	//assert(channels == 3 || channels == 4);
	int loc = (x * channels) + ((height - 1 - y) * stride);
	if (channels == 4)
	{
		return NColor8(pixels[loc + 2], pixels[loc + 1], pixels[loc], pixels[loc + 3]);
	}
	else if (channels == 3)
	{
		return NColor8(pixels[loc + 2], pixels[loc + 1], pixels[loc], 255);
	}
	else if (channels == 1)
	{
		if (palette)
		{
			RGBQUAD quad = palette[pixels[loc] % NumColorsInPalette];		// mod for safety
			return NColor8(quad.rgbRed, quad.rgbGreen, quad.rgbBlue, 255);
		}
		else
		{
			return NColor8(pixels[loc], pixels[loc], pixels[loc], 255);
		}
	}
	else
	{
		return NColor8(0,0,0,0);
	}
}

// This only works for 24-bit or 32-bit images
void NImage::setPixel(int x, int y, NColor8 color)
{
	// This may slow stuff down, but I'm sick of getting crashes here...
	if (x < 0 || x >= width || y < 0 || y >= height)
	{
		return;
	}

	//assert(channels == 3 || channels == 4);
	if (channels == 3 || channels == 4)
	{
		int loc = (x * channels) + ((height - 1 - y) * stride);
		if (channels == 4)
		{
			pixels[loc + 3] = color.a;
		}
		pixels[loc + 2] = color.r;
		pixels[loc + 1] = color.g;
		pixels[loc] = color.b;
	}
	else if (channels == 1)
	{
		setPixel(x, y, color.r);
	}
}

// This only works for 8-bit images
void NImage::setPixel(int x, int y, uchar value)
{
	assert(channels == 1);
	int loc = (x) + ((height - 1 - y) * stride);
	pixels[loc] = value;
}

bool NImage::init(int w, int h, int chan, int bpp)
{
	hdc = CreateCompatibleDC(NULL);

	// Default, 'cause whatevs.
	NumColorsInPalette = 256;

	if (bpp > 0)
	{
		// OMG NEW PATH TO SUPPORT 16-COLOR BITMAPS!!
		// Having channel as an input to this function is convenient and I don't want to break that,
		// but when biBitCount is less than 8, I get ZERO CHANNELS. :|

		if (bpp < 8)
		{
			chan = 1;

			NumColorsInPalette = int(pow(2.0f, float(bpp)));
		}
		else
		{
			// Do nothing. This is redundant.
			assert(chan == bpp / 8);
		}
	}

	width = w;
	height = h;
	channels = chan;

	if (bpp > 0 && bpp < 8)
	{
		// If we're doing 16-color images, stride needs to compensate for half-byte lookups
		opt_assert(channels == 1);
		stride = (width * bpp) / 8;

		while (stride % 4 != 0)
		{
			stride++;
		}

		stride = stride * 8 / bpp;
	}
	else
	{
		stride = width * channels;

		while (stride % 4 != 0)
		{
			stride++;
		}
	}

	BITMAPINFO info = {0};

	info.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	info.bmiHeader.biWidth = width;
	info.bmiHeader.biHeight = height;
	info.bmiHeader.biPlanes = 1;
	info.bmiHeader.biBitCount = ushort(channels * 8);
	info.bmiHeader.biCompression = BI_RGB;

	// set up default file header and info header stuff

	fileheader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	if (channels == 1)
	{
		fileheader.bfOffBits += NumColorsInPalette * sizeof(RGBQUAD);
	}
	fileheader.bfReserved1 = 0;
	fileheader.bfReserved2 = 0;
	fileheader.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + stride * height;
	fileheader.bfType = 19778;

	infoheader.biBitCount = ushort(channels * 8);
	infoheader.biClrImportant = 0;
	infoheader.biClrUsed = 0;
	infoheader.biCompression = BI_RGB;
	infoheader.biHeight = height;
	infoheader.biPlanes = 1;
	infoheader.biSize = sizeof(BITMAPINFOHEADER);
	infoheader.biSizeImage = stride * height;
	infoheader.biWidth = width;
	infoheader.biXPelsPerMeter = 0;
	infoheader.biYPelsPerMeter = 0;

	bmp = CreateDIBSection(hdc, &info, DIB_RGB_COLORS, (void**)&pixels, 0, 0);

	if (channels == 1)
	{
		if (palette)
		{
			delete[] palette;
		}
		palette = new RGBQUAD[NumColorsInPalette];
		memset(palette, 0, NumColorsInPalette * sizeof(RGBQUAD));
	}

	if (!bmp)
	{
		return false;
	}

	oldbmp = (HBITMAP)SelectObject(hdc, bmp);

	if (!oldbmp)
	{
		return false;
	}

	return true;
}

bool NBlitter::Blit(NImage& dest, const NImage& src, int dest_x, int dest_y, int src_x, int src_y, int w, int h, bool doScissorChecks)
{
	if (doScissorChecks)
	{
		// check width and height
		if (w <= 0 || h <= 0 || src.getChannels() != dest.getChannels())
			return false;

		// check source bitmap
		if (src_x < 0 || src_x + w > src.getWidth() || src_y < 0 || src_y + h > src.getHeight())
			return false;

		// check destination bitmap
		if (dest_x <= -w || dest_x >= dest.getWidth() || dest_y <= -h || dest_y >= dest.getHeight())
			return false;

		if (dest_x < 0)
		{
			src_x -= dest_x;
			w += dest_x;
			dest_x = 0;
		}
		if (dest_x + w > dest.getWidth())
		{
			w = dest.getWidth() - dest_x;
		}
		if (dest_y < 0)
		{
			src_y -= dest_y;
			h += dest_y;
			dest_y = 0;
		}
		if (dest_y + h > dest.getHeight())
		{
			h = dest.getHeight() - dest_y;
		}
	}

	// blitting is done one row at a time

	unsigned char* dest_offset = dest.pixels + (dest.getHeight() - (1 + dest_y)) * dest.getStride() + (dest_x * dest.getChannels());
	unsigned char* src_offset = src.pixels + (src.getHeight() - (1 + src_y)) * src.getStride() + (src_x * src.getChannels());

	int size = sizeof(unsigned char) * (w * Max(src.getChannels(),dest.getChannels()));

	for (int j = 0; j < h; ++j)
	{
		memcpy(dest_offset, src_offset, size);
		dest_offset -= dest.getStride();
		src_offset -= src.getStride();
	}

	return true;
}

bool NBlitter::MaskedBlit(NImage& dest, const NImage& src, int dest_x, int dest_y, int src_x, int src_y, int w, int h, bool doScissorChecks)
{
	if (doScissorChecks)
	{
		// check width and height
		if (w <= 0 || h <= 0)
			return false;

		// check source bitmap
		if (src_x < 0 || src_x + w > src.getWidth() || src_y < 0 || src_y + h > src.getHeight())
			return false;

		// check destination bitmap
		if (dest_x <= -w || dest_x >= dest.getWidth() || dest_y <= -h || dest_y >= dest.getHeight())
			return false;

		if (dest_x < 0)
		{
			src_x -= dest_x;
			w += dest_x;
			dest_x = 0;
		}
		if (dest_x + w > dest.getWidth())
		{
			w = dest.getWidth() - dest_x;
		}
		if (dest_y < 0)
		{
			src_y -= dest_y;
			h += dest_y;
			dest_y = 0;
		}
		if (dest_y + h > dest.getHeight())
		{
			h = dest.getHeight() - dest_y;
		}
	}

	// blitting has to be done one pixel at a time to check for masking

	unsigned char* dest_offset = dest.pixels + (dest.getHeight() - (1 + dest_y)) * dest.getStride() + (dest_x * dest.getChannels());
	unsigned char* src_offset = src.pixels + (src.getHeight() - (1 + src_y)) * src.getStride() + (src_x * src.getChannels());
	int row_offset = (w * dest.getChannels());
	int dest_row_offset = dest.getStride() + row_offset;
	int src_row_offset = src.getStride() + row_offset;

	int size = sizeof(unsigned char) * Max(src.getChannels(),dest.getChannels());

	// 07/01/2010: TODO: Optimize this by blitting whole chunks of opaque pixels at once.
	// Still have to READ every pixel from the source image, but at least the writes can be done all at once.

	for (int j = src_y; j < src_y + h; ++j)
	{
		for (int i = src_x; i < src_x + w; ++i)
		{
			if (src.getPixel(i, j).ToARGB() != src.maskColor.ToARGB())
			{
				memcpy(dest_offset, src_offset, size);
			}
			dest_offset += dest.getChannels();
			src_offset += src.getChannels();
		}
		dest_offset -= dest_row_offset;
		src_offset -= src_row_offset;
	}

	return true;
}

bool NBlitter::TransparentBlit(NImage& dest, const NImage& src, int dest_x, int dest_y, int src_x, int src_y, int w, int h, bool doScissorChecks, float opacity)
{
	if (doScissorChecks)
	{
		// check width, height, and opacity
		if (w <= 0 || h <= 0 || opacity < 0.0f || opacity > 1.0f)
			return false;

		// check source bitmap
		if (src_x < 0 || src_x + w > src.getWidth() || src_y < 0 || src_y + h > src.getHeight())
			return false;

		// check destination bitmap
		if (dest_x <= -w || dest_x >= dest.getWidth() || dest_y <= -h || dest_y >= dest.getHeight())
			return false;

		if (dest_x < 0)
		{
			src_x -= dest_x;
			w += dest_x;
			dest_x = 0;
		}
		if (dest_x + w > dest.getWidth())
		{
			w = dest.getWidth() - dest_x;
		}
		if (dest_y < 0)
		{
			src_y -= dest_y;
			h += dest_y;
			dest_y = 0;
		}
		if (dest_y + h > dest.getHeight())
		{
			h = dest.getHeight() - dest_y;
		}
	}

	// blitting has to be done one pixel at a time to check for masking and to calculate blending

	unsigned char* dest_offset = dest.pixels + (dest.getHeight() - (1 + dest_y)) * dest.getStride() + (dest_x * dest.getChannels());
	unsigned char* src_offset = src.pixels + (src.getHeight() - (1 + src_y)) * src.getStride() + (src_x * src.getChannels());
	int s_row_offset = (w * src.getChannels());
	int d_row_offset = (w * dest.getChannels());
	int dest_row_offset = dest.getStride() + s_row_offset;
	int src_row_offset = src.getStride() + d_row_offset;

	if (!IsEqual(opacity,0.5f,0.001f))
	{

		float transparency = 1.0f - opacity;

		int size = sizeof(unsigned char);

		for (int j = src_y; j < src_y + h; ++j)
		{
			for (int i = src_x; i < src_x + w; ++i)
			{
				if (src.getPixel(i, j).ToARGB() != src.maskColor.ToARGB())
				{
					unsigned char bBlend = (unsigned char)(float(*dest_offset * transparency) + float(*src_offset * opacity));
					unsigned char gBlend = (unsigned char)(float(*(dest_offset + 1) * transparency) + float(*(src_offset + 1) * opacity));
					unsigned char rBlend = (unsigned char)(float(*(dest_offset + 2) * transparency) + float(*(src_offset + 2) * opacity));

					memcpy(dest_offset, (char*)&bBlend, size);
					memcpy(dest_offset + 1, (char*)&gBlend, size);
					memcpy(dest_offset + 2, (char*)&rBlend, size);
				}
				dest_offset += dest.getChannels();
				src_offset += src.getChannels();
			}
			dest_offset -= dest_row_offset;
			src_offset -= src_row_offset;
		}

	}
	else
	{

		//float transparency = 1.0f - opacity;

		int size = sizeof(unsigned char);

		for (int j = src_y; j < src_y + h; ++j)
		{
			for (int i = src_x; i < src_x + w; ++i)
			{
				if (src.getPixel(i, j).ToARGB() != src.maskColor.ToARGB())
				{
					unsigned char bBlend = (unsigned char)((*dest_offset + *src_offset) >> 1);
					unsigned char gBlend = (unsigned char)((*(dest_offset+1) + *(src_offset+1)) >> 1);
					unsigned char rBlend = (unsigned char)((*(dest_offset+2) + *(src_offset+2)) >> 1);

					memcpy(dest_offset, (char*)&bBlend, size);
					memcpy(dest_offset + 1, (char*)&gBlend, size);
					memcpy(dest_offset + 2, (char*)&rBlend, size);
				}
				dest_offset += dest.getChannels();
				src_offset += src.getChannels();
			}
			dest_offset -= dest_row_offset;
			src_offset -= src_row_offset;
		}

	}

	return true;
}

bool NBlitter::AlphaBlit(NImage& dest, const NImage& src, int dest_x, int dest_y, int src_x, int src_y, int w, int h, bool doScissorChecks)
{
	if (doScissorChecks)
	{
		// check width and height
		if (w <= 0 || h <= 0 || src.getChannels() != 4)
			return false;

		// check source bitmap
		if (src_x < 0 || src_x + w > src.getWidth() || src_y < 0 || src_y + h > src.getHeight())
			return false;

		// check destination bitmap
		if (dest_x <= -w || dest_x >= dest.getWidth() || dest_y <= -h || dest_y >= dest.getHeight())
			return false;

		if (dest_x < 0)
		{
			src_x -= dest_x;
			w += dest_x;
			dest_x = 0;
		}
		if (dest_x + w > dest.getWidth())
		{
			w = dest.getWidth() - dest_x;
		}
		if (dest_y < 0)
		{
			src_y -= dest_y;
			h += dest_y;
			dest_y = 0;
		}
		if (dest_y + h > dest.getHeight())
		{
			h = dest.getHeight() - dest_y;
		}
	}

	// blitting has to be done one pixel at a time to check for masking

	unsigned char* dest_offset = dest.pixels + (dest.getHeight() - (1 + dest_y)) * dest.getStride() + (dest_x * dest.getChannels());
	unsigned char* src_offset = src.pixels + (src.getHeight() - (1 + src_y)) * src.getStride() + (src_x * src.getChannels());
	int s_row_offset = (w * src.getChannels());
	int d_row_offset = (w * dest.getChannels());
	int dest_row_offset = dest.getStride() + d_row_offset;
	int src_row_offset = src.getStride() + s_row_offset;

	int size = sizeof(unsigned char);

	for (int j = src_y; j < src_y + h; ++j)
	{
		for (int i = src_x; i < src_x + w; ++i)
		{
			if (src.getPixel(i, j).ToARGB() != src.maskColor.ToARGB())
			{
				unsigned char bBlend = (unsigned char)((*(src_offset + 3) * (*(src_offset) - *(dest_offset)) + (*(dest_offset)<<8)) >> 8);
				unsigned char gBlend = (unsigned char)((*(src_offset + 3) * (*(src_offset+1) - *(dest_offset+1)) + (*(dest_offset+1)<<8)) >> 8);
				unsigned char rBlend = (unsigned char)((*(src_offset + 3) * (*(src_offset+2) - *(dest_offset+2)) + (*(dest_offset+2)<<8)) >> 8);

				memcpy(dest_offset, (char*)&bBlend, size);
				memcpy(dest_offset + 1, (char*)&gBlend, size);
				memcpy(dest_offset + 2, (char*)&rBlend, size);
			}
			dest_offset += dest.getChannels();
			src_offset += src.getChannels();
		}
		dest_offset -= dest_row_offset;
		src_offset -= src_row_offset;
	}

	return true;
}

bool NBlitter::ScaleDownBlit(NImage& dest, const NImage& src, int dest_x, int dest_y, int src_x, int src_y, int dest_w, int dest_h, int src_w, int src_h, bool doScissorChecks, bool filter)
{
	if (doScissorChecks)
	{
	}

	// these will be greater than one if scaling down
	float xScale = float(src_w) / float(dest_w);
	float yScale = float(src_h) / float(dest_h);

	for (int x = dest_x; x < dest_x + dest_w; ++x)
	{
		for (int y = dest_y; y < dest_y + dest_h; ++y)
		{
			// for each pixel in the destination image, sum the weighted pixels from the source image
			float a = 0.0f;
			float r = 0.0f;
			float g = 0.0f;
			float b = 0.0f;

			float f_src_x_start = float(src_x) + (float(x - dest_x) * xScale);
			float f_src_x_end = float(src_x) + (float(x+1 - dest_x) * xScale);
			float f_src_y_start = float(src_y) + (float(y - dest_y) * yScale);
			float f_src_y_end = float(src_y) + (float(y+1 - dest_y) * yScale);

			int src_x_start = int(f_src_x_start);
			int src_x_end = filter ? Max(int(f_src_x_end), src_x_start + 1) : src_x_start + 1;
			int src_y_start = int(f_src_y_start);
			int src_y_end = filter ? Max(int(f_src_y_end), src_y_start + 1) : src_y_start + 1;

			// if src_x_end is an integer, it means the source width ends BEFORE that pixel;
			// a one-pixel source segment WILL have different start and end points

			float totalWeight = 0.0f;

			for (int _x = src_x_start; _x < src_x_end; ++_x)
			{
				for (int _y = src_y_start; _y < src_y_end; ++_y)
				{
					NColor8 srcPixel = src.getPixel(_x, _y);

					float c_f_src_x_start = Max(f_src_x_start, float(int(f_src_x_start)));
					float c_f_src_x_end = Max(f_src_x_end, float(int(f_src_x_end)));
					float c_f_src_y_start = Max(f_src_y_start, float(int(f_src_y_start)));
					float c_f_src_y_end = Max(f_src_y_end, float(int(f_src_y_end)));

					float xWidth = Min(1.0f, c_f_src_x_end - c_f_src_x_start);
					float yWidth = Min(1.0f, c_f_src_y_end - c_f_src_y_start);

					float xWeight = 1.0f / (xWidth);
					float yWeight = 1.0f / (yWidth);
					float weight = xWeight * yWeight;

					totalWeight += weight;

					a += (float(srcPixel.a) * weight);
					r += (float(srcPixel.r) * weight);
					g += (float(srcPixel.g) * weight);
					b += (float(srcPixel.b) * weight);
				}
			}

			//NLog(Text("totalWeight(%d,%d) = %f", x, y, totalWeight));

			a /= totalWeight;
			r /= totalWeight;
			g /= totalWeight;
			b /= totalWeight;

			dest.setPixel(x, y, NColor8(uchar(r), uchar(g), uchar(b), uchar(a)));
		}
	}

	return true;
}

bool NBlitter::ScaleUpBlit(NImage& dest, const NImage& src, int dest_x, int dest_y, int src_x, int src_y, int dest_w, int dest_h, int src_w, int src_h, bool doScissorChecks, bool filter)
{
	if (doScissorChecks)
	{
		// check width and height
		if (dest_w <= 0 || dest_h <= 0 || src_w <= 0 || src_h <= 0 || src.getChannels() != dest.getChannels())
			return false;

		// check source bitmap
		if (src_x < 0 || src_x + src_w > src.getWidth() || src_y < 0 || src_y + src_h > src.getHeight())
			return false;

		// check destination bitmap
		if (dest_x < -dest_w || dest_x > dest.getWidth() || dest_y < -dest_h || dest_y > dest.getHeight())
			return false;
	}

	// blitting has to be done one pixel at a time to scale
	float delta_width = float(src_w) / float(dest_w);
	float delta_height = float(src_h) / float(dest_h);

	for (int j = dest_y; j < dest_y + dest_h; ++j)
	{
		for (int i = dest_x; i < dest_x + dest_w; ++i)
		{
			NColor8 final;

			if (filter)
			{

				float scaledSrcX = Max(float(i - dest_x) * delta_width - 0.5f, float(src_x));
				float scaledSrcY = Max(float(j - dest_y) * delta_height - 0.5f, float(src_y));

				int src_i1 = src_x + int(scaledSrcX);
				int src_j1 = src_y + int(scaledSrcY);

				int src_i2 = Min(src_i1 + 1, src_x + src_w - 1);
				int src_j2 = Min(src_j1 + 1, src_y + src_h - 1);

				NColor8 pix1 = src.getPixel(src_i1, src_j1);
				NColor8 pix2 = src.getPixel(src_i2, src_j1);
				NColor8 pix3 = src.getPixel(src_i1, src_j2);
				NColor8 pix4 = src.getPixel(src_i2, src_j2);

				float xBias = scaledSrcX - float(int(scaledSrcX));
				float yBias = scaledSrcY - float(int(scaledSrcY));

				NColor8 topBlend = NColor8(uchar((float(pix1.r) * (1.0f - xBias)) + (float(pix2.r) * xBias)),
										  uchar((float(pix1.g) * (1.0f - xBias)) + (float(pix2.g) * xBias)),
										  uchar((float(pix1.b) * (1.0f - xBias)) + (float(pix2.b) * xBias)),
										  uchar((float(pix1.a) * (1.0f - xBias)) + (float(pix2.a) * xBias)));

				NColor8 bottomBlend = NColor8(uchar((float(pix3.r) * (1.0f - xBias)) + (float(pix4.r) * xBias)),
											 uchar((float(pix3.g) * (1.0f - xBias)) + (float(pix4.g) * xBias)),
											 uchar((float(pix3.b) * (1.0f - xBias)) + (float(pix4.b) * xBias)),
											 uchar((float(pix3.a) * (1.0f - xBias)) + (float(pix4.a) * xBias)));

				final = NColor8(uchar((float(topBlend.r) * (1.0f - yBias)) + (float(bottomBlend.r) * yBias)),
							   uchar((float(topBlend.g) * (1.0f - yBias)) + (float(bottomBlend.g) * yBias)),
							   uchar((float(topBlend.b) * (1.0f - yBias)) + (float(bottomBlend.b) * yBias)),
							   uchar((float(topBlend.a) * (1.0f - yBias)) + (float(bottomBlend.a) * yBias)));

			}
			else
			{

				float scaledSrcX = float(i - dest_x) * delta_width;// - 0.5f;
				float scaledSrcY = float(j - dest_y) * delta_height;// - 0.5f;

				int src_i1 = src_x + int(scaledSrcX);
				int src_j1 = src_y + int(scaledSrcY);

				final = src.getPixel(src_i1, src_j1);

			}

			if (i >= 0 && i < dest.getWidth() &&
				j >= 0 && j < dest.getHeight())
			{
				dest.setPixel(i, j, final);		// This is slow.  A memcpy would be faster,
												// but I am not intending to use this scaling function
												// in a real-time situation, and this is easier.
			}
		}
	}

	return true;
}

void NImage::clear()
{
	for (int i = 0; i < width; ++i)
	{
		for (int j = 0; j < height; ++j)
		{
			setPixel(i, j, maskColor);
		}
	}
}

void NImage::free()
{
	width = 0;
	height = 0;
	channels = 0;
	stride = 0;

	SelectObject(hdc,oldbmp);
	DeleteObject(bmp);
	ReleaseDC(NULL,hdc);
	DeleteDC(hdc);
	DeleteObject(oldbmp);

	SafeDeleteArray(palette);
}
