#pragma once

// This is a row-major 4x4 matrix class done without loops.

// Add NVec4 functions!!

// NERD: Do I need or want matrix classes besides 4x4?

#include "NMathDefs.h"
#include "NMathFtns.h"
#include "NVec2.h"
#include "NVec3.h"
#include "NVec4.h"

class NMat4;

// non-member functions to generate transformation matrices
NMat4 CreateScaleMatrix(const NVec3& s);
NMat4 CreateScaleMatrix(float s);
NMat4 CreateRotationMatrixX(float theta);
NMat4 CreateRotationMatrixY(float theta);
NMat4 CreateRotationMatrixZ(float theta);
NMat4 CreateRotationMatrix(const NVec3& n, float theta);
NMat4 CreateRotationMatrix(const NVec3& x, const NVec3& y, const NVec3& z);
NMat4 CreateProjectionMatrix(const NVec3& n);
NMat4 CreateReflectionMatrix(const NVec3& n);
NMat4 CreateTranslationMatrix(const NVec3& d);

NMat4 CreateProjMatrix(float fov, float aspectRatio, float nearClip, float farClip, float targetAspectRatio = -1.0f);
NMat4 CreateViewMatrix(const NVec3& pos, const NVec3& forward, const NVec3& left, const NVec3& up);
NMat4 CreateOrthoMatrix(float left, float right, float top, float bottom, float near, float far);

class NMat4
{
public:
	// constructors
	NMat4();										// Default constructor; returns the identity matrix
	NMat4(float inval[16]);							// Takes a 4x4 array of floats as parameters for new matrix values

	// operators
	NMat4 operator +(const NMat4& m) const;			// Matrix addition
	NMat4 operator -(const NMat4& m) const;			// Matrix subtraction
	NMat4 operator *(const NMat4& m) const;			// Matrix-matrix multiplication
	NVec2 operator *(const NVec2& v) const;			// Matrix-vector multiplication
	NVec3 operator *(const NVec3& v) const;			// Matrix-vector multiplication
	NVec4 operator *(const NVec4& v) const;			// Matrix-vector multiplication

	// Since I am assuming column vectors, matrix-vector multiplication must be done in this order.
	// Multiplying a vector on the left by a matrix on the right is not allowed.

	NMat4 operator *(float s) const;				// Matrix scalar multiplcation
	NMat4 operator /(float s) const;				// Matrix scalar division

	void operator +=(const NMat4& m);				// Matrix addition
	void operator -=(const NMat4& m);				// Matrix subtraction
	void operator *=(const NMat4& m);				// Matrix-matrix multiplication
	void operator *=(float s);						// Matrix scalar multiplcation
	void operator /=(float s);						// Matrix scalar division

	// member functions to return new matrices
	NMat4 getTranspose() const;						// return the transpose of this matrix
	NMat4 getInverse() const;							// return the inverse of this matrix

	NMat4 scaleBy(const NVec3& s) const {return CreateScaleMatrix(s) * (*this);}
													// return this matrix scaled by the given vector
	NMat4 scaleBy(float s) const {return CreateScaleMatrix(s) * (*this);}
													// return this matrix scaled by the given vector

	NMat4 rotateXBy(float theta) const {return CreateRotationMatrixX(theta) * (*this);}
	NMat4 rotateYBy(float theta) const {return CreateRotationMatrixY(theta) * (*this);}		
	NMat4 rotateZBy(float theta) const {return CreateRotationMatrixZ(theta) * (*this);}
	NMat4 rotateBy(const NVec3& n, float theta) const {return CreateRotationMatrix(n, theta) * (*this);}
													// return this vector rotated by the given angle
													// (in radians) around a cardinal axis
													// or around an arbitrary vector N

	NMat4 projectOnto(const NVec3& n) const {return CreateProjectionMatrix(n) * (*this);}
													// return this matrix projected onto a plane with normal N

	NMat4 reflectBy(const NVec3& n) const {return CreateReflectionMatrix(n) * (*this);}
													// return this matrix reflected about a plane with normal N

	NMat4 translateBy(const NVec3& d) const {return CreateTranslationMatrix(d) * (*this);}
													// return this matrix translated in the direction D

	float cofactor(int i, int j) const;				// get the specified cofactor
	float getDeterminant() const;					// get the determinant

	float val[16];									// these are the entries of the matrix
													// [0-3] are row 1, etc.
													// they are public because I don't see much danger in altering them

	void GetAxes(NVec3& x, NVec3& y, NVec3& z);
};

// non-member functions
// (These are all just inlined wrappers for the members functions defined above,
// in case this particular syntax is desired.)
inline NMat4 Transpose(const NMat4& m) {return m.getTranspose();}
inline NMat4 Invert(const NMat4& m) {return m.getInverse();}
inline NMat4 Scale(const NMat4& m, const NVec3& s) {return m.scaleBy(s);}
inline NMat4 Scale(const NMat4& m, float s) {return m.scaleBy(s);}
inline NMat4 RotateX(const NMat4& m, float theta) {return m.rotateXBy(theta);}
inline NMat4 RotateY(const NMat4& m, float theta) {return m.rotateYBy(theta);}
inline NMat4 RotateZ(const NMat4& m, float theta) {return m.rotateZBy(theta);}
inline NMat4 Rotate(const NMat4& m, const NVec3& n, float theta) {return m.rotateBy(n, theta);}
inline NMat4 Projection(const NMat4& m, const NVec3& n) {return m.projectOnto(n);}
inline NMat4 Reflection(const NMat4& m, const NVec3& n) {return m.reflectBy(n);}
inline NMat4 Translation(const NMat4& m, const NVec3& d) {return m.translateBy(d);}
