#pragma once

#include <math.h>
#include "NMathDefs.h"
#include "NMathFtns.h"

class NVec3;
class NVec4;
class NMat4;

class NVec2;
inline NVec2 Negate(const NVec2& v);

class NVec2
{
public:
	// Constructors
	NVec2() : x(0.0f), y(0.0f) {}
	NVec2(float _x) : x(_x), y(_x) {}
	NVec2(float _x, float _y) : x(_x), y(_y) {}
	NVec2(const NVec3& v);
	NVec2(const NVec4& v);

	// Member variables
	float x, y;

	// Overloaded operators
	NVec2 operator +(const NVec2& v) const	{return NVec2(x + v.x, y + v.y);}
	NVec2 operator -(const NVec2& v) const	{return NVec2(x - v.x, y - v.y);}
	NVec2 operator *(float s) const			{return NVec2(x * s, y * s);}
	NVec2 operator *(const NVec2& v) const	{return NVec2(x * v.x, y * v.y);}
	NVec2 operator /(float s) const			{float is = 1.0f / s; return NVec2(x * is, y * is);}
	NVec2 operator /(const NVec2& v) const	{return NVec2(x / v.x, y / v.y);}
	void operator +=(const NVec2& v)			{x += v.x; y += v.y;}
	void operator -=(const NVec2& v)			{x -= v.x; y -= v.y;}
	void operator *=(float s)				{x *= s; y *= s;}
	void operator *=(const NVec2& v)			{x *= v.x; y *= v.y;}
	void operator /=(float s)				{float is = 1.0f / s; x *= is; y *= is;}
	void operator /=(const NVec2& v)			{x /= v.x; y /= v.y;}

	NVec2 operator *(const NMat4& m) const;
	void operator *=(const NMat4& m);

	// Member functions
	float length() const					{return sqrt(lengthSquared());}
	float lengthSquared() const				{return x*x + y*y;}
	void normalize()						{float il = 1.0f / length(); il = (il < INFINITY) ? il : 0.0f; x *= il; y *= il;}
	void fastNormalize()
	{
		float il = FastInvSqrt((x*x)+(y*y));
		x *= il; y *= il;
	}
	void negate()							{x = -x; y = -y;}
	NVec2 operator -() const {return Negate(*this);}

private:
};

// Additional functions
inline NVec2 operator *(float s, const NVec2& v) {return v * s;}

inline float Dot(const NVec2& v, const NVec2& w) {return ((v.x * w.x) + (v.y * w.y));}
inline float Cross(const NVec2& v, const NVec2& w) {return v.x * w.y - v.y * w.x;}
inline NVec2 Normalize(const NVec2& v) {float il = 1.0f / v.length(); il = (il < INFINITY) ? il : 0.0f; return NVec2(v.x * il, v.y * il);}
inline NVec2 FastNormalize(const NVec2& v)
{
	float il = FastInvSqrt((v.x*v.x)+(v.y*v.y));
	return NVec2(v.x * il, v.y * il);
}
inline NVec2 Projection(const NVec2& dir, const NVec2& onto) {if (onto.lengthSquared() > 0.0f) {return onto * (Dot(dir, onto) / Dot(onto, onto));} else {return dir;}}
inline NVec2 Negate(const NVec2& v) {return NVec2(-v.x, -v.y);}

inline NVec2 Lerp(const NVec2& a, const NVec2& b, float t)
{
	return NVec2(Lerp(a.x, b.x, t),
				Lerp(a.y, b.y, t));
}

inline NVec2 Smooth(const NVec2& a, const NVec2& b, float speed, float deltaTime)
{
	return NVec2(Smooth(a.x, b.x, speed, deltaTime),
				Smooth(a.y, b.y, speed, deltaTime));
}
